package com.laguerre.kan;

import java.util.ArrayList;
import java.util.Arrays;

import com.laguerre.kan.SceneConductor.ISurfaceChangedListener;
import android.opengl.GLES20;
import android.opengl.Matrix;

public final class Camera extends Component implements ISurfaceChangedListener {

	/**
	 * Contains values that are invariant for a {@link Camera} that renders
	 * objects, such as the view matrix, the projection matrix, the concatenation
	 * of the view and the projection matrix, or the eye/camera position.
	 */
	public static final class RenderingValues {
		private final int[] viewSize = new int[2];
		private final Vector3 eyePosition = Vector3.getNewZeroVec();
		private final float[] viewMatrix = new float[16];
		private final float[] invViewMatrix = new float[16];
		private final float[] projMatrix = new float[16];
		private final float[] invProjMatrix = new float[16];
		private final float[] viewProjMatrix = new float[16];
		private final float[] invViewProjMatrix = new float[16];

		RenderingValues() {
			Arrays.fill(this.viewSize, 100);
			Matrix.setIdentityM(this.viewMatrix, 0);
			Matrix.setIdentityM(this.invViewMatrix, 0);
			Matrix.setIdentityM(this.projMatrix, 0);
			Matrix.setIdentityM(this.invProjMatrix, 0);
			Matrix.setIdentityM(this.viewProjMatrix, 0);
			Matrix.setIdentityM(this.invViewProjMatrix, 0);
		}

		/**
		 * Sets the size of the view in pixels.
		 * @param width
		 * @param height
		 */
		public void setViewSize(int width, int height) {
			this.viewSize[0] = width;
			this.viewSize[1] = height;
		}

		/**
		 * Sets the size of the view in pixels.
		 * @param viewSize where viewSize[0] contains width and viewSize[1]
		 *   contains height.
		 */
		public void setViewSize(int[] viewSize) {
			System.arraycopy(viewSize, 0, this.viewSize, 0, 2);
		}

		/**
		 * Sets width of the view in pixels.
		 * @param width
		 */
		public void setViewSizeWidth(int width) {
			this.viewSize[0] = width;
		}

		/**
		 * Sets height of the view in pixels.
		 * @param width
		 */
		public void setViewSizeHeight(int height) {
			this.viewSize[1] = height;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains view size in pixels in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the view size.
		 */
		public int[] getSharedViewSize() {
			return this.viewSize;
		}

		/**
		 * Sets the eye/camera position in world space.
		 * @param position
		 */
		public void setEyePosition(Vector3 position) {
			this.eyePosition.setValues(position);
		}

		/**
		 * Gets a direct reference to the inner <code>Vector3</code> that
		 * contains eye position in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the eye position.
		 */
		public Vector3 getSharedEyePosition() {
			return this.eyePosition;
		}

		/**
		 * Sets the view matrix and computes its inverse automatically.
		 * @param viewMatrix
		 */
		public void setViewMatrix(float[] viewMatrix) {
			setViewMatrix(viewMatrix, null);
		}

		/**
		 * Sets the view matrix and its inverse. If the inverse equals to
		 * <code>null</code>, it will be computed automatically.
		 * @param viewMatrix the view matrix.
		 * @param invViewMatrix the inverse of the view matrix or <code>null</code>.
		 */
		public void setViewMatrix(float[] viewMatrix, float[] invViewMatrix) {
			setMatrixes(viewMatrix, this.viewMatrix, invViewMatrix, this.invViewMatrix);
		}

		/**
		 * Sets the view-projection matrix and computes its inverse automatically.
		 * @param projViewMatrix
		 */
		public void setViewProjMatrix(float[] projViewMatrix) {
			setViewProjMatrix(projViewMatrix, null);
		}

		/**
		 * Sets the view-projection matrix and its inverse. If the inverse equals
		 * to <code>null</code>, it will be computed automatically.
		 * @param viewProjMatrix the view-projection matrix.
		 * @param invViewProjMatrix the inverse of the view-projection matrix or <code>null</code>.
		 */
		public void setViewProjMatrix(float[] viewProjMatrix, float[] invViewProjMatrix) {
			setMatrixes(viewProjMatrix, this.viewProjMatrix, invViewProjMatrix, this.invViewProjMatrix);
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the view matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the view matrix.
		 */
		public float[] getSharedViewMatrix() {
			return this.viewMatrix;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the inverse view matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the inverse view matrix.
		 */
		public float[] getSharedInvViewMatrix() {
			return this.invViewMatrix;
		}

		/**
		 * Sets the projection matrix and computes its inverse automatically.
		 * @param projMatrix
		 */
		public void setProjMatrix(float[] projMatrix) {
			setProjMatrix(projMatrix, null);
		}

		/**
		 * Sets the projection matrix and its inverse. If the inverse equals
		 * to <code>null</code>, it will be computed automatically.
		 * @param projMatrix the projection matrix.
		 * @param invProjMatrix the inverse of the projection matrix or <code>null</code>.
		 */
		public void setProjMatrix(float[] projMatrix, float[] invProjMatrix) {
			setMatrixes(projMatrix, this.projMatrix, invProjMatrix, this.invProjMatrix);
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the projection matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the projection matrix.
		 */
		public float[] getSharedProjMatrix() {
			return this.projMatrix;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the inverse projection matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the inverse projection matrix.
		 */
		public float[] getSharedInvProjMatrix() {
			return this.invProjMatrix;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the view-projection matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the view-projection matrix.
		 */
		public float[] getSharedViewProjMatrix() {
			return this.viewProjMatrix;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the inverse view-projection matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the inverse view-projection matrix.
		 */
		public float[] getSharedInvViewProjMatrix() {
			return this.invViewProjMatrix;
		}

		/**
		 * Fills a <code>float[]</code> array that represents a matrix and also
		 * set its inverse: to the proposed value or automatically if the value
		 * is <code>null</code>.
		 * @param src the matrix to copy
		 * @param fieldDst the reference to the array that will contain the copy of <code>src</code>.
		 * @param invSrc the inverse of the matrix to copy, or <code>null</code> to compute it automatically.
		 * @param fieldInvDst the reference to the array that will contain the copy of the inverse matrix.
		 */
		private void setMatrixes(float[] src, float[] fieldDst, float[] invSrc, float[] fieldInvDst) {
			System.arraycopy(src, 0, fieldDst, 0, src.length);
			if (invSrc != null) {
				System.arraycopy(invSrc, 0, fieldInvDst, 0, src.length);
			} else {
				Matrix.invertM(fieldInvDst, 0, src, 0);
			}
		}
	}

	private static Camera currentCamera = null;
	public static Camera getCurrent() {
		return currentCamera;
	}

	private boolean enabled;
	private RenderTexture renderTexture;
	private final ImageEffects postEffects = new ImageEffects();
	private int fbHandle;
	private int depthFbHandle;
	private boolean glParamsInitialized = false;
	private int passNum;
	private boolean needsPass;
	private final ArrayList<SceneObject> multiPassSOb = new ArrayList<SceneObject>();

	/**
	 * The default value of the culling mask is <code>1&lt;&lt;0</code>.
	 */
	private int cullingMask = 1 << 0;

	private final float[] backgroundColor = new float[4];
	private float fovY = 45f;
	private float nearClipPlane = 0.01f;
	private float farClipPlane = 20f;

	private final float[] viewMatrix = new float[16];
	private final float[] projMatrix = new float[16];
	private final float[] viewProjMatrix = new float[16];

	private final RenderingValues renderingPrms = new RenderingValues();

	Camera() {
		this.enabled = true;
		this.renderTexture = null;
		this.fbHandle = -1;
		this.depthFbHandle = -1;

		setBackgroundColor(0.3f, 0.3f, 0.3f, 1.0f);
		// Defaults to rendering to texture.
		this.postEffects.setRenderingMode(ImageEffects.FINAL_PASS_TO_TEXTURE);
	}

	public void setEnabled(boolean enabling) {
		this.enabled = enabling;
	}

	public boolean isEnabled() {
		return this.enabled;
	}

	public void setRenderTexture(RenderTexture renderTexture) {
		if (this.renderTexture != renderTexture) {
			this.renderTexture = renderTexture;
			setupFBO();
			updateProj();
			if (!this.postEffects.isInitialized()) {
				this.postEffects.init();
			}
			this.postEffects.setRenderTexture(renderTexture);
		}
	}

	private void setupFBO() {
		// Deletes the previous FBO, if any.
		freeResources();

		final int[] names = { -1 };
		// FB
		GLES20.glGenFramebuffers(1, names, 0);
		this.fbHandle = names[0];
		// depth
		GLES20.glGenRenderbuffers(1, names, 0);
		this.depthFbHandle = names[0];
		GLES20.glBindRenderbuffer(GLES20.GL_RENDERBUFFER, this.depthFbHandle);
		GLES20.glRenderbufferStorage(
				GLES20.GL_RENDERBUFFER, GLES20.GL_DEPTH_COMPONENT16,
				this.renderTexture.getWidth(), this.renderTexture.getHeight());
		GLES20.glBindRenderbuffer(GLES20.GL_RENDERBUFFER, 0);
	}

	/**
	 * Free resources used by this camera: the frame buffer and render buffers
	 * are deleted.
	 */
	public void freeResources() {
		final int[] rendBuff = new int[] { this.fbHandle };
		if (this.fbHandle > 0) {
			GLES20.glDeleteFramebuffers(1, rendBuff, 0);
			this.fbHandle = -1;
		}
		if (this.depthFbHandle > 0) {
			rendBuff[0] = this.depthFbHandle;
			GLES20.glDeleteRenderbuffers(1, rendBuff, 0);
			this.depthFbHandle = 1;
		}
		// Same thing for post effects.
		this.postEffects.freeResources();
	}

	public RenderTexture getSharedRenderTexture() {
		return this.renderTexture;
	}

	public void setAsRenderingOnScreen() {
		if (currentCamera != null) {
			currentCamera.postEffects.setRenderingMode(ImageEffects.FINAL_PASS_TO_TEXTURE);
			SceneConductor.getInstance().removeSurfaceChangedListener(currentCamera);
		}
		currentCamera = this;
		this.postEffects.setRenderingMode(ImageEffects.FINAL_PASS_TO_SCREEN);
		SceneConductor.getInstance().addSurfaceChangedListener(this);
	}

	public boolean isRenderingOnScreen() {
		return this == currentCamera;
	}

	/**
	 * Sets the culling mask. The default value of the culling mask is
	 * <code>1&lt;&lt;0</code>.
	 * @param mask
	 * @see getCullingMask
	 * @see isMaskCulled
	 */
	public void setCullingMask(int mask) {
		this.cullingMask = mask;
	}

	/**
	 * Gets the culling mask.
	 * @return the value of the culling mask.
	 * @see setCullingMask
	 * @see isMaskCulled
	 */
	public int getCullingMask() {
		return this.cullingMask;
	}

	/**
	 * If the bitwise <code>AND</code> operation between this mask and a
	 * {@link SceneObject} layer returns <code>0</code>, then the
	 * {@link SceneObject} will not be rendered.
	 * @see setCullingMask
	 * @see getCullingMask
	 */
	public boolean isMaskCulled(SceneObject sob) {
		return (sob.getLayer() & this.cullingMask) == 0;
	}

	@Override
	public void onSurfaceChanged(int width, int height) {
		updateProj();
	}

	public int getWidth() {
		return this.renderTexture.getWidth();
	}

	public int getHeight() {
		return this.renderTexture.getHeight();
	}

	public float[] getProjMatrix(float[] projMatrix) {
		System.arraycopy(this.projMatrix, 0, projMatrix, 0, 16);
		return projMatrix;
	}

	protected void updateProj() {
		final int width = getWidth();
		final int height = getHeight();

		this.renderingPrms.setViewSize(width, height);

		float ratio =  ((float) width) / height;
		perspectiveM(this.projMatrix, 0, this.fovY, ratio, this.nearClipPlane, this.farClipPlane);

		// Sets the projection matrix.
		this.renderingPrms.setProjMatrix(this.projMatrix);
	}

	public void setBackgroundColor(float r, float g, float b, float a) {
		this.backgroundColor[0] = r;
		this.backgroundColor[1] = g;
		this.backgroundColor[2] = b;
		this.backgroundColor[3] = a;
	}

	protected void initGLParams() {
//		GLES20.glEnable(GLES20.GL_DEPTH_TEST);
//		GLES20.glClearDepthf(1.0f);
//		GLES20.glDepthFunc(GLES20.GL_LEQUAL);
//		GLES20.glDepthMask(true);

		// cull backface
		GLES20.glEnable(GLES20.GL_CULL_FACE);
		GLES20.glCullFace(GLES20.GL_BACK);

		this.glParamsInitialized = true;
	}

	/**
	 * When beginning a new cycle of rendering, this method let mark the current
	 * {@link RenderTexture} as not swapped (i.e. marked as not containing a
	 * previous rendering.
	 */
	public void emptyRenderTexture() {
		this.renderTexture.markAsNotSwapped();
	}

	public void setToFirstPass() {
		this.passNum = 0;
		this.needsPass = true;
	}

	public int getCurrentPass() {
		return this.passNum;
	}

	public boolean renderOnePassToTexture() {
		if (!this.needsPass) {
			return false;
		}

		preRenderToTexture();

		// will be updated once a rendering occurred.
		this.needsPass = false;

		// Sets the eye position (in word space)
		this.renderingPrms.setEyePosition(getSharedTransform().getSharedWorldTranslation());

		// Sets the view matrix (i.e. inverse of the camera transform).
		getSharedTransform().getInverseWorldMatrix4(this.viewMatrix);
		this.renderingPrms.setViewMatrix(this.viewMatrix);

		// Proj x view
		Matrix.multiplyMM(this.viewProjMatrix, 0, this.projMatrix, 0, this.viewMatrix, 0);
		this.renderingPrms.setViewProjMatrix(this.viewProjMatrix);

		final Camera currentCamera = this;

		if (this.passNum == 0) {
			// First pass (i.e. pass 0).
			SceneConductor.getInstance().visitEachSceneObject(new SceneConductor.ISceneObjectVisitor() {
				@Override
				public void visitSceneObject(SceneObject sob) {
					if (sob.isEnabled() && !isMaskCulled(sob)) {
						final Renderer renderer = sob.getSharedRenderer();
						if (renderer != null) {
							renderer.setToFirstPass();
							renderer.renderOnePass(currentCamera);
							if (renderer.stillNeedsPass()) {
								multiPassSOb.add(sob);
							}
						}
					}
				}
			});
			this.needsPass = (this.multiPassSOb.size() > 0);
		} else {
			final int nbSObs = this.multiPassSOb.size();
			for (int iSOb = 0; iSOb < nbSObs; iSOb++) {
				final SceneObject sob = this.multiPassSOb.get(iSOb);
				if (sob != null) {
					final Renderer renderer = sob.getSharedRenderer();
					renderer.renderOnePass(currentCamera);
					if (!renderer.stillNeedsPass()) {
						// This sob does not need other passes, so discard it.
						this.multiPassSOb.set(iSOb, null);
					} else {
						this.needsPass = true;
					}
				}
			}
		}

		this.renderTexture.swapPrevAndCurrentTexture();

		// Prepare for the next run.
		this.passNum++;

		if (!needsPass) {
			this.multiPassSOb.clear();
		}

		return this.needsPass;
	}

	public boolean stillNeedsPass() {
		return this.needsPass;
	}

	private void preRenderToTexture() {
		Texture.resetTextureCounter();

		if (this.fbHandle < 1) {
			throw new Utils.WrongFrameBufferHandleException(
					"FBO not initialized? (fbHandle = " + this.fbHandle + ")");
		}

		GLES20.glBindFramebuffer(GLES20.GL_FRAMEBUFFER, this.fbHandle);

		GLES20.glFramebufferTexture2D(
				GLES20.GL_FRAMEBUFFER, GLES20.GL_COLOR_ATTACHMENT0,
                GLES20.GL_TEXTURE_2D, this.renderTexture.getTextureHandle(), 0);

		GLES20.glFramebufferRenderbuffer(
				GLES20.GL_FRAMEBUFFER, GLES20.GL_DEPTH_ATTACHMENT,
				GLES20.GL_RENDERBUFFER, this.depthFbHandle);

		final int fbStatus = GLES20.glCheckFramebufferStatus(GLES20.GL_FRAMEBUFFER);
		if (fbStatus != GLES20.GL_FRAMEBUFFER_COMPLETE) {
			throw new Utils.FrameBufferIncompleteException(
					"Could not bind frame buffer: " + fbStatus);
		}

		if (!this.glParamsInitialized) {
			initGLParams();
		}

    	GLES20.glClear(GLES20.GL_COLOR_BUFFER_BIT | GLES20.GL_DEPTH_BUFFER_BIT);

		// Set the background frame color
		GLES20.glClearColor(
				this.backgroundColor[0],
				this.backgroundColor[1],
				this.backgroundColor[2],
				this.backgroundColor[3]);

		GLES20.glViewport(0, 0, this.renderTexture.getWidth(), this.renderTexture.getHeight());
	}

	/**
	 * Gets parameters that can be passed to all shaders for a rendering with
	 * this {@link Camera}. Note that info contained by the returned object will
	 * be updated <b>only</b> by the method <code>render</code>.
	 *
	 * @return an object that contains info for rendering.
	 */
	public RenderingValues getSharedRenderingValues() {
		return this.renderingPrms;
	}

	/**
     * Define a projection matrix in terms of a field of view angle, an
     * aspect ratio, and z clip planes.
     * This code has been extracted from Android API level 14.
     *
     * @param m the float array that holds the perspective matrix
     * @param offset the offset into float array m where the perspective
     * matrix data is written
     * @param fovy field of view in y direction, in degrees
     * @param aspect width to height aspect ratio of the viewport
     * @param zNear
     * @param zFar
     */
    public static void perspectiveM(float[] m, int offset,
          float fovy, float aspect, float zNear, float zFar) {
        float f = 1.0f / (float) Math.tan(fovy * (Math.PI / 360.0));
        float rangeReciprocal = 1.0f / (zNear - zFar);

        m[offset + 0] = f / aspect;
        m[offset + 1] = 0.0f;
        m[offset + 2] = 0.0f;
        m[offset + 3] = 0.0f;

        m[offset + 4] = 0.0f;
        m[offset + 5] = f;
        m[offset + 6] = 0.0f;
        m[offset + 7] = 0.0f;

        m[offset + 8] = 0.0f;
        m[offset + 9] = 0.0f;
        m[offset + 10] = (zFar + zNear) * rangeReciprocal;
        m[offset + 11] = -1.0f;

        m[offset + 12] = 0.0f;
        m[offset + 13] = 0.0f;
        m[offset + 14] = 2.0f * zFar * zNear * rangeReciprocal;
        m[offset + 15] = 0.0f;
    }

    public void useAlternativePostEffectShader(Shader shader) {
    	this.postEffects.useAlternativeShader(shader);
    }

    public void setToFirstPostEffectPass() {
    	this.postEffects.setToFirstPass();
    }

    public int getCurrentPostEffectPass() {
    	return this.postEffects.getCurrentPass();
    }

    public boolean renderOnePostEffectPass() {
    	return this.postEffects.renderOnePass();
    }

    public boolean postEffectStillNeedsPass() {
    	return this.postEffects.stillNeedsPass();
    }
}
