package com.laguerre.kan;

import android.util.SparseArray;

public final class Lights {
	private static Lights instance = null;
	public static Lights getInstance() {
		if (instance == null) {
			instance = new Lights();
		}
		return instance;
	}
	public static void resetInstance() {
		instance = null;
	}

	public static final int MAX_PER_PIXEL_LIGHTS = 2;
	private final SparseArray<SceneObject> lights = new SparseArray<SceneObject>();

	private Lights() {
	}

	public void addLight(SceneObject sob) {
		this.lights.put(sob.getId(), sob);
	}

	public void removeLight(SceneObject sob) {
		this.lights.remove(sob.getId());
	}

	public void removeLights() {
		this.lights.clear();
	}

	public int lightCount() {
		return this.lights.size();
	}

	private final Vector3 lightDirection = Vector3.getNewOneVec();

	/**
	 * Passes the following values for each known light (max: <code>MAX_PER_PIXEL_LIGHTS</code>)
	 * to the given <code>shader</code>:
	 * <ul>
	 * <li><code>lightType</code> with: 0 => Directional, 1 => Point, 2 => Spot</li>
	 * <li><code>position</code>, which is the light position in world space</li>
	 * <li><code>direction</code>, which is the negated direction of the light
	 *  in world space (used only with directional and spot lights)</li>
	 * <li><code>diffuseColor</code>, which is the diffuse color</li>
	 * <li><code>specularColor</code>, which is the specular color (used only
	 *  with point and specular lights)</li>
	 * <li><code>range</code>, which is a value to cut the light after this
	 *  distance (used only with spot lights)
	 * <li><code>cosInnerCone</code>, which is the cosine of the inner cut-off
	 *  angle of a spot light
	 * <li><code>cosOuterCone</code>, which is the cosine of the outer cut-off
	 *  angle of a spot light.
	 * </ul>
	 * If less than <code>MAX_PER_PIXEL_LIGHTS</code> lights have been provided,
	 * values of default lights will be passed: directional light with black colors,
	 * zero direction and zero range.
	 * If more than <code>MAX_PER_PIXEL_LIGHTS</code> lights have been provided,
	 * then only the first <code>MAX_PER_PIXEL_LIGHTS</code> lights will be used
	 * (the order is undefined).
	 *
	 * @param camera the current camera so lights from ignored layers are discarded.
	 * @param shader the shader that will receive the values.
	 * @see MAX_PER_PIXEL_LIGHTS
	 */
	public void passLightsToShader(Camera camera, Shader shader) {
		final int lightSize = this.lights.size();
		int lightNum = 0;
		for (int lightIndex = 0; lightIndex < lightSize && lightNum < MAX_PER_PIXEL_LIGHTS; lightIndex++) {
			final SceneObject sob = this.lights.valueAt(lightIndex);
			if (!camera.isMaskCulled(sob)) {
				final Light light = sob.getSharedLight();
				// The user is asked to imagine that the ray light goes forward but
				// we use "backward" ray to get an inverse a ray light for shaders.
				sob.getSharedTransform().getWorldRotation().rotateDirection(
						this.lightDirection, Vector3.BACKWARD);

				passLightToShader(
						shader, lightNum, light.getLightType(),
						sob.getSharedTransform().getWorldTranslation(),
						this.lightDirection, light.getSharedDiffuseColor(),
						light.getSharedSpecularColor(), light.getRange(),
						(float) Math.cos(Math.toRadians(light.getInnerCutOffAngle())),
						(float) Math.cos(Math.toRadians(light.getOuterCutOffAngle())));
				lightNum++;
			}
		}
		// Puts default values if needed.
		for ( ; lightNum < MAX_PER_PIXEL_LIGHTS; lightNum++) {
			passLightToShader(
					shader, lightNum, Light.DIRECTIONAL, Vector3.ZERO, Vector3.ZERO,
					Color.BLACK, Color.BLACK, 0.0f, 0.0f, 0.0f);
		}
	}

	private void passLightToShader(
			Shader shader, int lightNum, int lightType,
			Vector3 position, Vector3 direction, Color diffuseColor,
			Color specularColor, float range, float cosInnerCone, float cosOuterCone) {
		// Builds strings such as
		//  "lights[0].position", "lights[0].color", etc.
		// and passes associated values.
		final String indexStr = "[" + lightNum + "]";
		final String lightsStr = "lights" + indexStr + ".";
		shader.setShaderValue(lightsStr + "lightType", lightType);
		shader.setShaderVector(lightsStr + "position", position);
		shader.setShaderVector(lightsStr + "direction", direction);
		shader.setShaderVector(lightsStr + "diffuseColor", diffuseColor);
		shader.setShaderVector(lightsStr + "specularColor", specularColor);
		shader.setShaderValue(lightsStr + "range", range);
		shader.setShaderValue(lightsStr + "cosInnerCone", cosInnerCone);
		shader.setShaderValue(lightsStr + "cosOuterCone", cosOuterCone);
	}
}
