package com.laguerre.kan;

import java.nio.FloatBuffer;

import android.opengl.GLES20;
import android.util.Log;

public final class PassShader {
	private int programHandle = -1;

	/**
	 * See <code>Shader.addPassShader</code> to create a {@link PassShader}.
	 */
	PassShader() {
	}

	public final int useExistingProgram(int programHandle) {
		this.programHandle = programHandle;
		return this.programHandle;
	}

	public final int createProgram() {
		this.programHandle = GLES20.glCreateProgram();
		return this.programHandle;
	}

	public final int getProgramHandle() {
		return this.programHandle;
	}

	public final int attachVertexShader(String shaderCode) {
		return attachShader(GLES20.GL_VERTEX_SHADER, shaderCode);
	}

	public final int attachFragmentShader(String shaderCode) {
		return attachShader(GLES20.GL_FRAGMENT_SHADER, shaderCode);
	}

	public final int attachShader(int shaderType, String shaderCode) {
		final int handle = loadShader(shaderType, shaderCode);
		if (handle != 0) {
			GLES20.glAttachShader(this.programHandle, handle);
		}
		return handle;
	}

	public final int attachShader(int shaderHandle) {
		GLES20.glAttachShader(this.programHandle, shaderHandle);
		return shaderHandle;
	}

	protected static int loadVertexShader(String shaderCode) {
		return loadShader(GLES20.GL_VERTEX_SHADER, shaderCode);
	}

	protected static int loadFragmentShader(String shaderCode) {
		return loadShader(GLES20.GL_FRAGMENT_SHADER, shaderCode);
	}

	/**
	 * Loads source and compiles a vertex or fragment shader.
	 * @param shaderType - GLES20.GL_VERTEX_SHADER or GLES20.GL_FRAGMENT_SHADER.
	 * @param shaderCode - The string that contains the shader code.
	 * @return The handle of the loaded shader.
	 */
	public static int loadShader(int shaderType, String shaderCode) {
	    int shaderHandle = GLES20.glCreateShader(shaderType);
	    GLES20.glShaderSource(shaderHandle, shaderCode);
	    GLES20.glCompileShader(shaderHandle);

        if (!isShaderCompiled(shaderHandle)) {
            Log.e("kan", "Could not compile shader:\n" + getShaderInfoLog(shaderHandle));
            GLES20.glDeleteShader(shaderHandle);
            shaderHandle = 0;
        }
	    return shaderHandle;
	}

	public void detachShader(int shaderHandle) {
		GLES20.glDetachShader(this.programHandle, shaderHandle);
	}

	public static boolean isShaderCompiled(int shaderHandle) {
		final int[] compiled = { -1 };
		GLES20.glGetShaderiv(shaderHandle, GLES20.GL_COMPILE_STATUS, compiled, 0);
		return compiled[0] != 0;
	}

	public static String getShaderInfoLog(int shaderHandle) {
		return GLES20.glGetShaderInfoLog(shaderHandle);
	}

	public final boolean linkProgram() {
		boolean linked = true;
		GLES20.glLinkProgram(this.programHandle);
		if (!isProgramLinked()) {
			Log.e("kan", "Linking failed:\n" + getProgramInfoLog());
			linked = false;
		}
		return linked;
	}

	public boolean isProgramLinked() {
		final int[] linked = { -1 };
		GLES20.glGetProgramiv(this.programHandle, GLES20.GL_LINK_STATUS, linked, 0);
		return linked[0] != 0;
	}

	public String getProgramInfoLog() {
		return GLES20.glGetProgramInfoLog(this.programHandle);
	}

	public void destroyResources() {
		GLES20.glDeleteProgram(this.programHandle);
		this.programHandle = -1;
	}

	protected final int passFloatVertexArray(
			String variableName, int size, int stride, FloatBuffer data) {
		final int handle = GLES20.glGetAttribLocation(
    			this.programHandle, variableName);
		if (handle != -1) {
    		GLES20.glEnableVertexAttribArray(handle);
    		GLES20.glVertexAttribPointer(handle, size, GLES20.GL_FLOAT, false, stride, data);
		}
		return handle;
	}

	final int passUniformInt(String variableName, int value) {
		int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
			GLES20.glUniform1i(handle, value);
		}
		return handle;
	}

	final int passUniformFloat(String variableName, float value) {
		int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
			GLES20.glUniform1f(handle, value);
		}
		return handle;
	}

	final int passUniformVec(String variableName, int[] vec, int vecType) {
		int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
			switch (vecType) {
			case 1:
				GLES20.glUniform1i(handle, vec[0]);
				break;
			case 2:
				GLES20.glUniform2i(handle, vec[0], vec[1]);
				break;
			case 3:
				GLES20.glUniform3i(handle, vec[0], vec[1], vec[2]);
				break;
			case 4:
				GLES20.glUniform4i(handle, vec[0], vec[1], vec[2], vec[3]);
				break;
			}
		}
		return handle;
	}

	final int passUniformVector3(String variableName, Vector3 v) {
		int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
			GLES20.glUniform3f(handle, v.getX(), v.getY(), v.getZ());
		}
		return handle;
	}

	final int passUniformVec(String variableName, float[] vec, int vecType) {
		int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
			switch (vecType) {
			case 1:
				GLES20.glUniform1f(handle, vec[0]);
				break;
			case 2:
				GLES20.glUniform2f(handle, vec[0], vec[1]);
				break;
			case 3:
				GLES20.glUniform3f(handle, vec[0], vec[1], vec[2]);
				break;
			case 4:
				GLES20.glUniform4f(handle, vec[0], vec[1], vec[2], vec[3]);
				break;
			}
		}
		return handle;
	}

	final int passUniformMatrix(String variableName, float[] matrix, int matType) {
		int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
			switch (matType) {
			case 2:
				GLES20.glUniformMatrix2fv(handle, 1, false, matrix, 0);
				break;
			case 3:
				GLES20.glUniformMatrix3fv(handle, 1, false, matrix, 0);
				break;
			case 4:
				GLES20.glUniformMatrix4fv(handle, 1, false, matrix, 0);
				break;
			}
		}
		return handle;
	}

	final int passUniformTexture(String variableName, int textureNum, Texture texture) {
		return passUniformTextureHandle(variableName, textureNum, texture.getTextureHandle());
	}

	final int passUniformTextureHandle(String variableName, int textureNum, int textureHandle) {
		int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
		    GLES20.glActiveTexture(Texture.getGLTexNum(textureNum));
		    GLES20.glBindTexture(GLES20.GL_TEXTURE_2D, textureHandle);
		    GLES20.glUniform1i(handle, textureNum);
		}
		return handle;
	}
}
