package com.laguerre.kan;

import java.util.ArrayList;
import java.util.HashSet;

import android.util.SparseArray;

public final class SceneConductor {
	public interface ISceneObjectVisitor {
		public void visitSceneObject(SceneObject sob);
	}

	public interface ISurfaceChangedListener {
		public void onSurfaceChanged(int width, int height);
	}

	private static SceneConductor instance = null;
	private final SparseArray<SceneObject> sobs = new SparseArray<SceneObject>();
	private final HashSet<ISurfaceChangedListener> surfaceChangedListeners =
			new HashSet<ISurfaceChangedListener>();
	private final ArrayList<SceneObject> camsToRender = new ArrayList<SceneObject>();

	public static SceneConductor getInstance() {
		if (instance == null) {
			instance = new SceneConductor();
		}
		return instance;
	}
	public static void resetInstance() {
		instance = null;
	}

	private SceneConductor() {
	}

	public void onRenderSurfaceChanged(int screenWidth, int screenHeight) {
		for (ISurfaceChangedListener listener : this.surfaceChangedListeners) {
			listener.onSurfaceChanged(screenWidth, screenHeight);
		}
	}

	public void addSurfaceChangedListener(ISurfaceChangedListener listener) {
		this.surfaceChangedListeners.add(listener);
	}

	public void removeSurfaceChangedListener(ISurfaceChangedListener listener) {
		this.surfaceChangedListeners.remove(listener);
	}

	public SceneObject addSceneObject() {
		return addSceneObject((String) null);
	}

	public SceneObject addSceneObject(String sobName) {
		final SceneObject sob = new SceneObject();
		sob.setName(sobName);
		this.sobs.put(sob.getId(), sob);
		return sob;
	}

	public void destroySceneObject(SceneObject sob) {
		if (sob != null) {
			this.sobs.remove(sob.getId());
			sob.destroy();
		}
	}

	public void destroySceneObject(int sobId) {
		destroySceneObject(this.sobs.get(sobId));
	}

	final void addCamera(SceneObject camSob) {
		if (!this.camsToRender.contains(camSob)) {
			this.camsToRender.add(camSob);
		}
	}

	final void removeCamera(SceneObject camSob) {
		this.camsToRender.remove(camSob);
	}

	public void drawFrame() {
		final int sobsSize = this.sobs.size();
		// First update states,
		for (int i = 0; i < sobsSize; i++) {
			final SceneObject sob = this.sobs.valueAt(i);
			sob.doLifeStep();
		}

		// Then draw objects.
		//
		final Camera currentCam = Camera.getCurrent();
		// First all the cameras that render to textures in the scene.
		for (SceneObject camSob : this.camsToRender) {
			final Camera sharedCam = camSob.getSharedCamera();
			if (currentCam != sharedCam && sharedCam.isEnabled()) {
				renderCam(sharedCam);
			}
		}
		// Finally the main camera.
		renderCam(currentCam);
	}

	/**
	 * Renders all the passes of a {@link Camera} and then all the passes of its
	 * post-effects.
	 * @param cam the {@link Camera} to render.
	 */
	private void renderCam(Camera cam) {
		cam.emptyRenderTexture();
		cam.setToFirstPass();
		while (cam.stillNeedsPass()) {
			cam.getSharedSOb().onBeforeRenderingCameraPass(cam.getCurrentPass());
			cam.renderOnePassToTexture();
			Utils.checkGlError("Detection of OpenGL error after rendering");
		}
		cam.setToFirstPostEffectPass();
		while (cam.postEffectStillNeedsPass()) {
			cam.getSharedSOb().onBeforeRenderingImageEffectsPass(cam.getCurrentPostEffectPass());
			cam.renderOnePostEffectPass();
			Utils.checkGlError("Detection of OpenGL error after post-effect");
		}
	}

	public void visitEachSceneObject(ISceneObjectVisitor visitor) {
		final int sobsSize = this.sobs.size();
		for (int i = 0; i < sobsSize; i++) {
			SceneObject sob = this.sobs.valueAt(i);
			visitor.visitSceneObject(sob);
		}
	}
}
