package com.laguerre.kan;

import android.util.SparseArray;

public final class SceneObject extends BaseObject {
	private boolean enabled;
	private int layer;
	private final Transform transform = new Transform();
	private final SparseArray<Component> components = new SparseArray<Component>();
	private final SparseArray<Behaviour> behaviours = new SparseArray<Behaviour>();

	// Shortcuts
	private Renderer renderer = null;
	private Camera camera = null;
	private Light light = null;

	SceneObject() {
		this.enabled = true;
		this.layer = 1 << 0;
	}

	public SceneObject(SceneObject sob) {
		copySceneObjectContent(sob);
	}

	protected void copySceneObjectContent(SceneObject orig) {
		copyBaseObjectContent(orig);
		this.enabled = orig.enabled;
		this.transform.setValues(orig.transform);
	}

	public boolean isEnabled() {
		return this.enabled;
	}

	public void setEnabled(boolean enabled) {
		this.enabled = enabled;
		final int behaviourSize = this.behaviours.size();
		for(int i = 0; i < behaviourSize; i++) {
			final Behaviour behaviour = this.behaviours.valueAt(i);
			behaviour.setEnabled(enabled);
		}
	}

	public void setLayer(int layer) {
		this.layer = layer;
	}

	public int getLayer() {
		return this.layer;
	}

	public Transform getSharedTransform() {
		return this.transform;
	}

	public Renderer getSharedRenderer() {
		return this.renderer;
	}

	public Camera getSharedCamera() {
		return this.camera;
	}

	public Light getSharedLight() {
		return this.light;
	}

	public <T extends Component> T getAnyComponent(Class<T> type) {
		Component foundComponent = null;
		final int componentSize = this.components.size();
		for (int i = 0; i < componentSize && foundComponent == null; i++) {
			final Component component = this.components.valueAt(i);
			if (type.isAssignableFrom(component.getClass())) {
				foundComponent = component;
			}
		}
		return type.cast(foundComponent);
	}

	/**
	 * Updates the states of the {@link Behaviour} objects of this {@link SceneObject}
	 * if <code>isEnabled</code> returned <code>true</code>.
	 *
	 * @see SceneObject.isEnabled
	 */
	final void doLifeStep() {
		if (!isEnabled()) {
			return;
		}
		final int behaviourSize = this.behaviours.size();
		for(int i = 0; i < behaviourSize; i++) {
			final Behaviour behaviour = this.behaviours.valueAt(i);
			behaviour.doLifeStep();
		}
	}

	/**
	 * Called by a {@link Camera} just before it render one pass to dispatch
	 * this event to the {@link Behaviour} that is associated with the {@link Camera}.
	 * @param passNum
	 */
	final void onBeforeRenderingCameraPass(int passNum) {
		final int behaviourSize = this.behaviours.size();
		for(int i = 0; i < behaviourSize; i++) {
			final Behaviour behaviour = this.behaviours.valueAt(i);
			if (behaviour.isEnabled()) {
				behaviour.onBeforeRenderingCameraPass(passNum);
			}
		}
	}

	/**
	 * Called before one pass that will apply an image effect, to dispatch this
	 * event to the {@link Behaviour} that is associated with the {@link Camera}.
	 *
	 * The first image that will be provided is the result of the render passes
	 * of the current {@link Camera}.
	 *
	 * @param passNum
	 */
	public void onBeforeRenderingImageEffectsPass(int passNum) {
		final int behaviourSize = this.behaviours.size();
		for(int i = 0; i < behaviourSize; i++) {
			final Behaviour behaviour = this.behaviours.valueAt(i);
			if (behaviour.isEnabled()) {
				behaviour.onBeforeRenderingImageEffectsPass(passNum);
			}
		}
	}

	final void destroy() {
		final int behaviourSize = this.behaviours.size();
		for(int i = 0; i < behaviourSize; i++) {
			final Behaviour behaviour = this.behaviours.valueAt(i);
			behaviour.destroy();
		}
	}

	public <T extends Component> T addComponent(Class<T> componentClass) {
		return addComponent(componentClass, null);
	}

	public <T extends Component> T addComponent(Class<T> componentClass, String name) {
		T ret = null;
		try {
			ret = componentClass.newInstance();
			ret.setParentSceneObject(this);
			ret.setName(name);
			storeAddedComponent(ret);
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}
		return ret;
	}

	private <T extends Component> void storeAddedComponent(T component) {
		this.components.put(component.getId(), component);
		// if possible, store it also as SObBehaviour.
		if (component instanceof Behaviour) {
			final Behaviour behaviour = (Behaviour) component;
			this.behaviours.put(behaviour.getId(), behaviour);
		}
		// Is this component a shortcut that we handle?
		if (component instanceof Renderer) {
			if (this.renderer != null) {
				throw new DuplicatedComponent("Renderer already exist for this SOb.");
			}
			this.renderer = (Renderer) component;
		} else if (component instanceof Camera) {
			if (this.camera != null) {
				throw new DuplicatedComponent("Camera already exist for this SOb.");
			}
			this.camera = (Camera) component;
			// Make the camera "working".
			SceneConductor.getInstance().addCamera(this);
		} else if (component instanceof Light) {
			if (this.light != null) {
				throw new DuplicatedComponent("Light already exist for this SOb.");
			}
			this.light = (Light) component;
			Lights.getInstance().addLight(this);
		}
	}

	public <T extends Component> void removeComponent(T component) {
		this.components.remove(component.getId());
		if (component instanceof Behaviour) {
			this.behaviours.remove(component.getId());
		}
		// Is this component a shortcut that we handle?
		if (component instanceof Renderer) {
			this.renderer = null;
		} else if (component instanceof Camera) {
			this.camera = null;
			// Make the camera "not working".
			SceneConductor.getInstance().removeCamera(this);
		} else if (component instanceof Light) {
			Lights.getInstance().removeLight(this);
			this.light = null;
		}
	}

	public static class DuplicatedComponent extends RuntimeException {
		private static final long serialVersionUID = 7746453127776700693L;

		public DuplicatedComponent(String detailMessage) {
			super(detailMessage);
		}
	}
}
