package com.laguerre.kan;

import java.util.HashSet;

import android.opengl.Matrix;

public final class Transform {
	private Transform parent = null;
	private HashSet<Transform> children = new HashSet<Transform>();

	private final Vector3 worldTranslation = Vector3.getNewZeroVec();
	private final Quaternion worldRotation = Quaternion.getNewIdQuat();
	private final Vector3 worldScale= Vector3.getNewOneVec();

	private final Vector3 localTranslation = Vector3.getNewZeroVec();
	private final Quaternion localRotation = Quaternion.getNewIdQuat();
	private final Vector3 localScale= Vector3.getNewOneVec();

	private final float[] worldMatrix = new float[16];
	private boolean worldDirty = false;
	private final float[] localMatrix = new float[16];
	private boolean localDirty = false;
	private Transform cachedInverse = null;
	// set to true so it will be initialized before any use
	private boolean inverseDirty = true;

	/**
	 * Constant to compare against identity in order to avoid the creation of
	 * many intermediate objects.
	 * <br />
	 * Since Java cannot ensure constantness of the returned object, please do
	 * <b>not</b> modify it.
	 */
	public static final Transform IDENTITY = getNewIdentityTf();

	public static Transform getNewIdentityTf() {
		return new Transform();
	}

	public Transform() {
		this(Vector3.ZERO);
	}

	public Transform(Vector3 translation) {
		this(translation, Quaternion.QUATERNION_IDENTITY);
	}

	public Transform(Quaternion rotation) {
		this(Vector3.ZERO, rotation);
	}

	public Transform(Vector3 translation, Quaternion rotation) {
		this(translation, rotation, Vector3.ONE);
	}

	public Transform(Vector3 translation, Quaternion rotation, Vector3 scale) {
		setWorldValues(translation, rotation, scale);
	}

	public Transform(Transform t) {
		setValues(t);
	}

	public void setValues(Transform t) {
		if (t == this) {
			return;
		}
		this.parent = t.parent;
		// Replace children.
		this.children.clear();
		this.children.addAll(t.children);
		// World transform
		this.worldTranslation.setValues(t.worldTranslation);
		this.worldRotation.setValues(t.worldRotation);
		this.worldScale.setValues(t.worldScale);
		// Local transform
		this.localTranslation.setValues(t.localTranslation);
		this.localRotation.setValues(t.localRotation);
		this.localScale.setValues(t.localScale);
		// Matrixes and dirtiness
		System.arraycopy(t.worldMatrix, 0, this.worldMatrix, 0, t.worldMatrix.length);
		this.worldDirty = t.worldDirty;
		System.arraycopy(t.localMatrix, 0, this.localMatrix, 0, t.localMatrix.length);
		this.localDirty = t.localDirty;
		this.inverseDirty = t.inverseDirty;
		if (this.cachedInverse == null) {
			this.cachedInverse = getNewIdentityTf();
		}
		if (!t.inverseDirty) {
			// Avoids to copy the inverse matrix if it is needed to update its content.
			this.cachedInverse.setValues(t.cachedInverse);
		}
	}

	public void setWorldValues(Vector3 translation, Quaternion rotation, Vector3 scale) {
		setWorldValues(this.parent, translation, rotation, scale);
	}

	public void setLocalValues(Vector3 translation, Quaternion rotation, Vector3 scale) {
		setLocalValues(this.parent, translation, rotation, scale);
	}

	public void setWorldValues(Transform parent, Vector3 translation, Quaternion rotation, Vector3 scale) {
		setParent(parent);
		this.worldScale.setValues(scale);
		this.worldRotation.setValues(rotation);
		this.worldTranslation.setValues(translation);
		this.worldDirty = true;
		this.inverseDirty = true;
		updateLocalTransform();
	}

	public void setLocalValues(Transform parent, Vector3 translation, Quaternion rotation, Vector3 scale) {
		setParent(parent);
		this.localScale.setValues(scale);
		this.localRotation.setValues(rotation);
		this.localTranslation.setValues(translation);
		this.localDirty = true;
		this.inverseDirty = true;
		updateWorldTransformAndChildren();
	}

	@Override
	public String toString() {
		return "{{WT: " + this.worldTranslation +
				", WR: " + this.worldRotation +
				", WS: " + this.worldScale +
				"}, {lT: " + this.localTranslation +
				", lR: " + this.localRotation +
				", lS: " + this.localScale + "}}";
	}

	public boolean equalsIgnoringParent(Object obj) {
		return objectEquals(obj, true);
	}

	@Override
	public boolean equals(Object obj) {
		return objectEquals(obj, false);
	}

	private boolean objectEquals(Object obj, boolean ignoreParent) {
		if (this == obj) {
			return true;
		} else if (!(obj instanceof Transform)) {
			return false;
		} else {
			Transform t = (Transform) obj;
			boolean parentsEqual = (ignoreParent ? true : this.parent == t.parent);
			return parentsEqual &&
					// Since world and local representations are synchronized,
					// we compare only the world representation.
					this.worldScale.equals(t.worldScale) &&
					this.worldRotation.equals(t.worldRotation) &&
					this.worldTranslation.equals(t.worldTranslation);
		}
	}

	@Override
	public int hashCode() {
		// Ignores "parent" to avoid a up-traversing of all the tree in order to
		// get hashcodes until the root Transform.
		// So we may end up with two Transform that are different (according to
		// equals()) but with the same hashcode, which is fine:
		// http://docs.oracle.com/javase/6/docs/api/java/lang/Object.html#hashCode%28%29
		return this.worldScale.hashCode() ^ this.worldRotation.hashCode() ^ this.worldTranslation.hashCode();
	}

	public void getWorldMatrix4(float[] result) {
		if (this.worldDirty) {
			getLocalMatrix4(this.worldMatrix);
			if (hasParent()) {
				this.parent.getWorldMatrix4(result);
				Matrix.multiplyMM(this.worldMatrix, 0, result, 0, this.localMatrix, 0);
			}
			this.worldDirty = false;
		}
		System.arraycopy(this.worldMatrix, 0, result, 0, this.worldMatrix.length);
	}

	public void getLocalMatrix4(float[] result) {
		if (this.localDirty) {
			Matrix.setIdentityM(this.localMatrix, 0);
			//
			Matrix.scaleM(this.localMatrix, 0,
					this.localScale.getX(),
					this.localScale.getY(),
					this.localScale.getZ());
			//
			final float[] angleAxis = new float[4];
			this.localRotation.getAngleAxis(angleAxis);
			Matrix.setRotateM(this.localMatrix, 0,
					angleAxis[0], angleAxis[1], angleAxis[2], angleAxis[3]);
			//
			Matrix.translateM(this.localMatrix, 0,
					this.localTranslation.getX(),
					this.localTranslation.getY(),
					this.localTranslation.getZ());

			this.localDirty = false;
		}
		System.arraycopy(this.localMatrix, 0, result, 0, this.localMatrix.length);
	}

	public boolean hasParent() {
		return this.parent != null;
	}

	/**
	 * Sets a parent to this {@link Transform}. Note that cyclic graphs are not
	 * supported.
	 *
	 * @param parent - The parent of this {@link Transform}, which cannot be itself.
	 */
	public void setParent(Transform parent) {
		if (this.parent != parent) {
			// Detach from previous parent.
			if (this.parent != null) {
				this.parent.detachChild(this);
			}
			// Set and attach to the new parent
			this.parent = parent;
			if (this.parent != null) {
				this.parent.attachChild(this);
			}
		}
	}

	/**
	 * Gets the immediate parent of this {@link Transform}.
	 *
	 * @return the parent {@link Transform} or <code>null</code> if none.
	 */
	public Transform getSharedParent() {
		return this.parent;
	}

	/**
	 * Gets the top-most {@link Transform} in the hierarchy. This method will
	 * never return <code>null</code> but the {@link Transform} itself.
	 *
	 * @return the top-most {@link Transform} or itself.
	 */
	public Transform getRootSharedParent() {
		Transform root = this;
		while (root.hasParent()) {
			root = root.parent;
		}
		return root;
	}

	private void attachChild(Transform t) {
		children.add(t);
	}

	private void detachChild(Transform t) {
		children.remove(t);
	}

	/**
	 * Gets the immediate children of this {@link Transform}.
	 *
	 * @param result - the children.
	 * @return <code>result</code> (that contains the children).
	 */
	public Transform[] getChildren(Transform[] result) {
		this.children.toArray(result);
		return result;
	}

	/**
	 * Gets the immediate children of this {@link Transform}.
	 *
	 * @return an array with the children.
	 */
	public Transform[] getChildren() {
		return (Transform[]) this.children.toArray();
	}

	/**
	 * Tells whether a {@link Transform} is an immediate child of this {@link Transform}.
	 *
	 * @param t - The {@link Transform} to test.
	 * @return <code>true</code> if the given {@link Transform} is a child,
	 *   <code>false</code> otherwise.
	 */
	public boolean isChildOf(Transform t) {
		return t.children.contains(this);
	}

	/**
	 * Tells whether a {@link Transform} is a descendant of this {@link Transform}.
	 * It means that it may be a child through many intermediate parent {@link Transform}.
	 * Note that <code>this.isDescendantOf(this)</code> returns <code>false</code>.
	 *
	 * @param t - The {@link Transform} to test. The value <code>null</code> is
	 *   accepted since it is a {@link Transform} without any parent, but every
	 *   {@link Transform} is also descendant of <code>null</code>.
	 * @return <code>true</code> if the given {@link Transform} is a descendant,
	 *   <code>false</code> otherwise.
	 */
	public boolean isDescendantOf(Transform t) {
		Transform curr = this.parent;
		while (curr != t && curr != null) {
			curr = curr.parent;
		}
		return curr == t;
	}

	public Vector3 getWorldScale(Vector3 result) {
		result.setValues(this.worldScale);
		return result;
	}

	public Vector3 getWorldScale() {
		return new Vector3(this.worldScale);
	}

	/**
	 * Gets a direct reference to the inner <code>Vector3</code> that contains
	 * scale in order to avoid memory allocations. Please, do <b>not</b> update
	 * scale components by this way, use <code>scaleGlobally</code> for example.
	 * @return A reference to the scale.
	 */
	public Vector3 getSharedWorldScale() {
		return this.worldScale;
	}

	public Quaternion getWorldRotation(Quaternion result) {
		result.setValues(this.worldRotation);
		return result;
	}

	public Quaternion getWorldRotation() {
		return new Quaternion(this.worldRotation);
	}

	/**
	 * Gets a direct reference to the inner <code>Quaternion</code> that
	 * contains scale in order to avoid memory allocations. Please, do
	 * <b>not</b> update rotation components by this way, use
	 * <code>rotateGlobally</code> for example.
	 * @return A reference to the rotation.
	 */
	public Quaternion getSharedWorldRotation() {
		return this.worldRotation;
	}

	public Vector3 getWorldTranslation(Vector3 result) {
		result.setValues(this.worldTranslation);
		return result;
	}

	public Vector3 getWorldTranslation() {
		return new Vector3(this.worldTranslation);
	}

	/**
	 * Gets a direct reference to the inner <code>Vector3</code> that contains
	 * translation in order to avoid memory allocations. Please, do <b>not</b>
	 * update translation components by this way, use <code>translateGlobally</code>
	 * for example.
	 * @return A reference to the translation.
	 */
	public Vector3 getSharedWorldTranslation() {
		return this.worldTranslation;
	}

	public Vector3 getLocalScale(Vector3 result) {
		result.setValues(this.localScale);
		return result;
	}

	public Vector3 getLocalScale() {
		return new Vector3(this.localScale);
	}

	/**
	 * Gets a direct reference to the inner <code>Vector3</code> that contains
	 * scale in order to avoid memory allocations. Please, do <b>not</b> update
	 * scale components by this way, use <code>scaleLocally</code> for example.
	 * @return A reference to the scale.
	 */
	public Vector3 getSharedLocalScale() {
		return this.localScale;
	}

	public Quaternion getLocalRotation(Quaternion result) {
		result.setValues(this.localRotation);
		return result;
	}

	public Quaternion getLocalRotation() {
		return new Quaternion(this.localRotation);
	}

	/**
	 * Gets a direct reference to the inner <code>Quaternion</code> that
	 * contains scale in order to avoid memory allocations. Please, do
	 * <b>not</b> update rotation components by this way, use
	 * <code>rotateLocally</code> for example.
	 * @return A reference to the rotation.
	 */
	public Quaternion getSharedLocalRotation() {
		return this.localRotation;
	}

	public Vector3 getLocalTranslation(Vector3 result) {
		result.setValues(this.localTranslation);
		return result;
	}

	public Vector3 getLocalTranslation() {
		return new Vector3(this.localTranslation);
	}

	/**
	 * Gets a direct reference to the inner <code>Vector3</code> that contains
	 * translation in order to avoid memory allocations. Please, do <b>not</b>
	 * update translation components by this way, use <code>translateLocally</code>
	 * for example.
	 * @return A reference to the translation.
	 */
	public Vector3 getSharedLocalTranslation() {
		return this.localTranslation;
	}

	public void scaleGlobally(Vector3 scale) {
		scaleGlobally(scale.getX(), scale.getY(), scale.getZ());
	}

	public void scaleGlobally(float x, float y, float z) {
		this.worldScale.scaleSelf(x, y, z);
		this.worldDirty = true;
		this.inverseDirty = true;
		updateLocalTransform();
	}

	public void scaleGlobally(float f) {
		scaleGlobally(f, f, f);
	}

	public void rotateGlobally(Quaternion q) {
		this.worldRotation.mulSelf(q);
		this.worldDirty = true;
		this.inverseDirty = true;
		updateLocalTransform();
	}

	public void translateGlobally(Vector3 translation) {
		translateGlobally(
				translation.getX(),
				translation.getY(),
				translation.getZ());
	}

	public void translateGlobally(float dx, float dy, float dz) {
		this.worldTranslation.addSelf(dx, dy, dz);
		this.worldDirty = true;
		this.inverseDirty = true;
		updateLocalTransform();
	}

	public void scaleLocally(Vector3 scale) {
		scaleGlobally(scale.getX(), scale.getY(), scale.getZ());
	}

	public void scaleLocally(float x, float y, float z) {
		this.localScale.scaleSelf(x, y, z);
		this.localDirty = true;
		this.inverseDirty = true;
		updateWorldTransformAndChildren();
	}

	public void scaleLocally(float f) {
		scaleLocally(f, f, f);
	}

	public void rotateLocally(Quaternion q) {
		this.localRotation.mulSelf(q);
		this.localDirty = true;
		this.inverseDirty = true;
		updateWorldTransformAndChildren();
	}

	public void translateLocally(Vector3 translation) {
		translateLocally(
				translation.getX(),
				translation.getY(),
				translation.getZ());
	}

	public void translateLocally(float dx, float dy, float dz) {
		this.localTranslation.addSelf(dx, dy, dz);
		this.localDirty = true;
		this.inverseDirty = true;
		updateWorldTransformAndChildren();
	}

	/**
	 * Updates the world transform values from the local transform. This method
	 * should be called every time local coordinates have been modified.
	 * <br />
	 * The world transform is computed along the parent transform. If no parent
	 * has been provided, the world values will be the ones from the local
	 * transform.
	 * <br />
	 * In addition, the world transform values of the children will also
	 * be updated by calling this function on each of them.
	 */
	private void updateWorldTransformAndChildren() {
		if (hasParent()) {
			// s_world = s_world_parent * s_local
			this.worldScale.setValues(
					this.parent.worldScale.getX() * this.localScale.getX(),
					this.parent.worldScale.getY() * this.localScale.getY(),
					this.parent.worldScale.getZ() * this.localScale.getZ());
			// q_world = q_world_parent * q_local
			Quaternion.mul(this.worldRotation, this.parent.worldRotation, this.localRotation);
			// t_world = t_world_parent * t_local
			//  so with vectors => t_world_parent + t_local
			this.worldTranslation.setValues(
					this.parent.worldTranslation.getX() + this.localTranslation.getX(),
					this.parent.worldTranslation.getY() + this.localTranslation.getY(),
					this.parent.worldTranslation.getZ() + this.localTranslation.getZ());
		} else {
			this.worldScale.setValues(this.localScale);
			this.worldRotation.setValues(this.localRotation);
			this.worldTranslation.setValues(this.localTranslation);
		}
		// Updates children, if any.
		for (Transform child : this.children) {
			child.updateWorldTransformAndChildren();
		}
		this.worldDirty = true;
	}

	/**
	 * Updates the local transform values from the global transform. This method
	 * should be called every time world coordinates have been modified.
	 * <br />
	 * The local transform is computed along the parent transform. If no parent
	 * has been provided, the local values will be the ones from the world
	 * transform.
	 */
	private void updateLocalTransform() {
		if (hasParent()) {
			// s_local = s_world_parent^-1 * s_world
			this.localScale.setValues(
					this.worldScale.getX() / this.parent.worldScale.getX(),
					this.worldScale.getY() / this.parent.worldScale.getY(),
					this.worldScale.getZ() / this.parent.worldScale.getZ());
			// q_local = q_world_parent^-1 * q_world
			Quaternion.inverse(this.localRotation, this.parent.worldRotation);
			this.localRotation.mulSelf(this.worldRotation);
			// t_local = t_world_parent^-1 * t_world
			//  so with vectors => t_world - t_world_parent
			this.localTranslation.setValues(
					this.worldTranslation.getX() - this.parent.worldTranslation.getX(),
					this.worldTranslation.getY() - this.parent.worldTranslation.getY(),
					this.worldTranslation.getZ() - this.parent.worldTranslation.getZ());
		} else {
			this.localScale.setValues(this.worldScale);
			this.localRotation.setValues(this.worldRotation);
			this.localTranslation.setValues(this.worldTranslation);
		}
		this.localDirty = true;
	}

	private void updateCachedInverse() {
		if (this.cachedInverse == null) {
			this.cachedInverse = getNewIdentityTf();
		}
		this.cachedInverse.worldScale.setValues(
				1.0f / this.worldScale.getX(),
				1.0f / this.worldScale.getY(),
				1.0f / this.worldScale.getZ());
		Quaternion.inverse(this.cachedInverse.worldRotation, this.worldRotation);
		Vector3.mul(this.cachedInverse.worldTranslation, -1.0f, this.worldTranslation);
		this.cachedInverse.worldDirty = true;
		// local
		this.cachedInverse.localScale.setValues(
				1.0f / this.localScale.getX(),
				1.0f / this.localScale.getY(),
				1.0f / this.localScale.getZ());
		Quaternion.inverse(this.cachedInverse.localRotation, this.localRotation);
		Vector3.mul(this.cachedInverse.localTranslation, -1.0f, this.localTranslation);
		this.cachedInverse.localDirty = true;
		// Say that we do not need another update now.
		this.inverseDirty = false;
	}

	public Transform getInverse(Transform result) {
		if (this.inverseDirty) {
			updateCachedInverse();
		}
		result.setValues(this.cachedInverse);
		return result;
	}

	public Transform getInverse() {
		Transform ret = new Transform();
		return getInverse(ret);
	}

	public void getInverseWorldMatrix4(float[] result) {
		if (this.inverseDirty) {
			updateCachedInverse();
		}
		this.cachedInverse.getWorldMatrix4(result);
	}

	public void getInverseLocalMatrix4(float[] result) {
		if (this.inverseDirty) {
			updateCachedInverse();
		}
		this.cachedInverse.getLocalMatrix4(result);
	}

	public float[] getInverse(float[] result, float[] matrix) {
		Matrix.invertM(result, 0, matrix, 0);
		return result;
	}

	public float[] getInverse(float[] matrix) {
		final float[] ret = new float[16];
		return getInverse(ret);
	}

	public static Transform combine(Transform result, Transform lT, Transform rT) {
		Vector3.scale(result.worldScale, lT.worldScale, rT.worldScale);
		Quaternion.mul(result.worldRotation, lT.worldRotation, rT.worldRotation);
		Vector3.add(result.worldTranslation, lT.worldTranslation, rT.worldTranslation);
		result.updateLocalTransform();
		return result;
	}

	public static float[] combine(float[] result, Transform lT, float[] rM) {
		Matrix.multiplyMM(result, 0, lT.worldMatrix, 0, rM, 0);
		return result;
	}

	public static float[] combine(Transform lT, float[] rM) {
		final float[] ret = new float[16];
		return combine(ret, lT, rM);
	}

	public static float[] combine(float[] result, float[] lM, Transform rT) {
		Matrix.multiplyMM(result, 0, lM, 0, rT.worldMatrix, 0);
		return result;
	}

	public static float[] combine(float[] lM, Transform rT) {
		final float[] ret = new float[16];
		return combine(ret, lM, rT);
	}

	public static float[] combine(float[] result, float[] lM, float[] rM) {
		Matrix.multiplyMM(result, 0, lM, 0, rM, 0);
		return result;
	}

	public static float[] combine(float[] lM, float[] rM) {
		final float[] ret = new float[16];
		return combine(ret, lM, rM);
	}

	public Vector3 transformDirectionToWorld(Vector3 v) {
		this.worldRotation.rotateDirection(v, v);
		return v;
	}

	public Vector3 transformDirectionToLocal(Vector3 v) {
		Quaternion worldInv = Quaternion.inverse(this.worldRotation);
		worldInv.rotateDirection(v, v);
		return v;
	}

	public Vector3 transformPointToWorld(Vector3 v) {
		Vector3.scale(v, this.worldScale, v);
		this.worldRotation.rotateDirection(v, v);
		return v;
	}

	public Vector3 transformPointToLocal(Vector3 v) {
		Vector3 invScale = new Vector3(
				1.0f / this.worldScale.getX(),
				1.0f / this.worldScale.getY(),
				1.0f / this.worldScale.getZ());
		Vector3.scale(v, invScale, v);
		Quaternion worldInv = Quaternion.inverse(this.worldRotation);
		worldInv.rotateDirection(v, v);
		return v;
	}
}
