package com.laguerre.mytest;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;

import com.laguerre.kan.ManualMesh;

public class Cube extends ManualMesh {
	private static final float[] cubeVertices = {
		/*
		 *   4+------+ 7      ^ y
		 *   /      / |       |
		 *  0+-----+3 |       0---> x
		 *   |     |  +6     z
		 *   |     | /
		 *  1+-----+2
		 *
		 */

		// front: 0, 1, 2, 0, 2, 3
		//  => 0, 1, 2, 0, 2, 3
		-0.5f,  0.5f,  0.5f, 1.0f,   // 0 - 0
		-0.5f, -0.5f,  0.5f, 1.0f,   // 1 - 1
		 0.5f, -0.5f,  0.5f, 1.0f,   // 2 - 2
		 0.5f,  0.5f,  0.5f, 1.0f,   // 3 - 3

		// right: 3, 2, 6, 3, 6, 7
		//  => 4, 5, 6, 4, 6, 7
		 0.5f,  0.5f,  0.5f, 1.0f,   // 3 - 4
		 0.5f, -0.5f,  0.5f, 1.0f,   // 2 - 5
		 0.5f, -0.5f, -0.5f, 1.0f,   // 6 - 6
		 0.5f,  0.5f, -0.5f, 1.0f,   // 7 - 7

		// back: 7, 6, 5, 7, 5, 4
		//  => 8, 9, 10, 8, 10, 11
		 0.5f,  0.5f, -0.5f, 1.0f,   // 7 - 8
		 0.5f, -0.5f, -0.5f, 1.0f,   // 6 - 9
		-0.5f, -0.5f, -0.5f, 1.0f,   // 5 - 10
		-0.5f,  0.5f, -0.5f, 1.0f,   // 4 - 11

		// left: 4, 5, 1, 4, 1, 0
		//  => 12, 13, 14, 12, 14, 15
		-0.5f,  0.5f, -0.5f, 1.0f,   // 4 - 12
		-0.5f, -0.5f, -0.5f, 1.0f,   // 5 - 13
		-0.5f, -0.5f,  0.5f, 1.0f,   // 1 - 14
		-0.5f,  0.5f,  0.5f, 1.0f,   // 0 - 15

		// top: 4, 0, 3, 4, 3, 7
		//  => 16, 17, 18, 16, 18, 19
		-0.5f,  0.5f, -0.5f, 1.0f,   // 4 - 16
		-0.5f,  0.5f,  0.5f, 1.0f,   // 0 - 17
		 0.5f,  0.5f,  0.5f, 1.0f,   // 3 - 18
		 0.5f,  0.5f, -0.5f, 1.0f,   // 7 - 19

		// bottom: 1, 5, 6, 1, 6, 2
		//  => 20, 21, 22, 20, 22, 23
		-0.5f, -0.5f,  0.5f, 1.0f,   // 1 - 20
		-0.5f, -0.5f, -0.5f, 1.0f,   // 5 - 21
		 0.5f, -0.5f, -0.5f, 1.0f,   // 6 - 22
		 0.5f, -0.5f,  0.5f, 1.0f,   // 2 - 23
	};

	private static final float[] normalCoords = {
		// front
		 0f,  0f,  1f,    0f,  0f,  1f,    0f,  0f,  1f,    0f,  0f,  1f,
		// right
		 1f,  0f,  0f,    1f,  0f,  0f,    1f,  0f,  0f,    1f,  0f,  0f,
		// back
		 0f,  0f, -1f,    0f,  0f, -1f,    0f,  0f, -1f,    0f,  0f, -1f,
		// left
		-1f,  0f,  0f,   -1f,  0f,  0f,   -1f,  0f,  0f,   -1f,  0f,  0f,
		// top
		 0f,  1f,  0f,    0f,  1f,  0f,    0f,  1f,  0f,    0f,  1f,  0f,
		// bottom
		 0f, -1f,  0f,    0f, -1f,  0f,    0f, -1f,  0f,    0f, -1f,  0f
	};

	private static short[] drawList = {
		// front
		0, 1, 2, 0, 2, 3,
		// right
		4, 5, 6, 4, 6, 7,
		// back
		8, 9, 10, 8, 10, 11,
		// left
		12, 13, 14, 12, 14, 15,
		// top
		16, 17, 18, 16, 18, 19,
		// bottom
		20, 21, 22, 20, 22, 23
	};

	private static final float[] cubeUV = {
		/*
		 *  Faces in OpenGL texture:
		 *
		 *  +---+---+---+
		 *  |fr |ri |ba +
		 *  +---+---+---+
		 *  |le |to |bo |
		 *  +---+---+---+
		 *
		 */
		// front
		0.0f, 1.0f,       0.0f, 0.5f,       0.333333f, 0.5f,   0.333333f, 1.0f,
		// right
		0.333333f, 1.0f,  0.333333f, 0.5f,  0.666666f, 0.5f,   0.666666f, 1.0f,
		// back
		0.666666f, 1.0f,  0.666666f, 1.0f,  1.0f, 0.5f,        1.0f, 1.0f,
		// left
		0.0f, 0.5f,       0.0f, 0.0f,       0.333333f, 0.0f,   0.333333f, 0.5f,
		// top
		0.333333f, 0.5f,  0.333333f, 0.0f,  0.666666f, 0.0f,   0.666666f, 0.5f,
		// bottom
		0.666666f, 0.5f,  0.666666f, 0.5f,  1.0f, 0.0f,        1.0f, 0.5f
	};

	private FloatBuffer vertexBuffer;
	private final int vertexStride = 4 * Float.SIZE / 8;
	private FloatBuffer normalBuffer;
    private final int normalStride = 3 * Float.SIZE / 8;;
	private FloatBuffer uvBuffer;
	private final int uvStride = 2 * Float.SIZE / 8;
	private ShortBuffer drawListBuffer;

    @Override
    public int getNbOfSubMeshes() {
    	return 1;
    }

	@Override
	public void build() {
    	ByteBuffer bb = ByteBuffer.allocateDirect(cubeVertices.length * Float.SIZE / 8);
    	bb.order(ByteOrder.nativeOrder());
    	this.vertexBuffer = bb.asFloatBuffer();
    	this.vertexBuffer.put(cubeVertices);
    	this.vertexBuffer.position(0);

        bb = ByteBuffer.allocateDirect(normalCoords.length * Float.SIZE / 8);
        bb.order(ByteOrder.nativeOrder());
        normalBuffer = bb.asFloatBuffer();
        normalBuffer.put(normalCoords);
        normalBuffer.position(0);

    	bb = ByteBuffer.allocateDirect(cubeUV.length * Float.SIZE / 8);
    	bb.order(ByteOrder.nativeOrder());
    	this.uvBuffer = bb.asFloatBuffer();
    	this.uvBuffer.put(cubeUV);
    	this.uvBuffer.position(0);

    	bb = ByteBuffer.allocateDirect(drawList.length * Short.SIZE / 8);
    	bb.order(ByteOrder.nativeOrder());
    	this.drawListBuffer = bb.asShortBuffer();
    	this.drawListBuffer.put(drawList);
    	this.drawListBuffer.position(0);
	}

	@Override
	public int getVertexCount(int subMeshId) {
		return cubeVertices.length;
	}

	@Override
	public int getVertexStride(int subMeshId) {
		return vertexStride;
	}

	@Override
	public FloatBuffer getSharedVertexBuffer(int subMeshId) {
		return this.vertexBuffer;
	}

	@Override
	public int getNormalStride(int subMeshId) {
		return this.normalStride;
	}

	@Override
	public FloatBuffer getSharedNormalBuffer(int subMeshId) {
		return this.normalBuffer;
	}

	@Override
	public int getSharedTexCoordStride(int subMeshId, int textureId) {
		return uvStride;
	}

	@Override
	public FloatBuffer getSharedTexCoordBuffer(int subMeshId, int textureId) {
		return this.uvBuffer;
	}

	@Override
	public int getDrawListCount(int subMeshId) {
		return drawList.length;
	}

	@Override
	public ShortBuffer getSharedDrawListBuffer(int subMeshId) {
		return this.drawListBuffer;
	}
}
